# -*- coding: utf-8 -*-
#
#       __init__.py
#
#       Copyright 2015 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

''' A simple and powerful dual-screen PDF reader designed for presentations.
'''

#
# DON'T IMPORT ANYTHING HERE (OR YOU WILL BREAK setup.py)
#

__version__ = '1.3.1'
__author__ = '''2009, 2010 Thomas Jost <thomas.jost@gmail.com>
2015-2019 Cimbali <me@cimba.li>
2016 Christoph Rath <christof.rath@iaik.tugraz.at>
2016 Epithumia <endless@airelle.info>
'''

__all__ = ['builder', 'config', 'document', 'editable_label', 'extras', 'media_overlay', 'pointer', 'scribble', 'surfacecache', 'talk_time', 'ui', 'util']
